import logging
import abc

import numpy as np
import torch
import typing

from arm_pytorch_utilities.tensor_utils import ensure_tensor
from torch.distributions import MultivariateNormal

from pytorch_mppi import MPPI
# optimizers
import cma

logger = logging.getLogger(__file__)


class EvaluationResult(typing.NamedTuple):
    # (N) cost for each trajectory evaluated
    costs: torch.Tensor
    # (N x H x nx) where H is the horizon and nx is the state dimension
    rollouts: torch.Tensor
    # parameter values populated by the tuner after evaluation returns
    params: dict = None
    # iteration number populated by the tuner after evaluation returns
    iteration: int = None


class Optimizer:
    def __init__(self):
        self.tuner: typing.Optional[Autotune] = None
        self.optim = None

    @abc.abstractmethod
    def setup_optimization(self) -> None:
        """Create backend optim object with optimization parameters and MPPI parameters from the tuner"""

    @abc.abstractmethod
    def optimize_step(self) -> EvaluationResult:
        """Optimize a single step, returning the evaluation result from the latest parameters"""

    def optimize_all(self, iterations) -> EvaluationResult:
        """Optimize multiple steps, returning the best evaluation results.
        Some optimizers may only have this implemented."""
        res = None
        for i in range(iterations):
            res = self.optimize_step()
        return res


class CMAESOpt(Optimizer):
    """Optimize using CMA-ES, an evolutionary algorithm that maintains a Gaussian population,
    starting around the initial parameters with a variance (potentially different for each hyperparameter)."""

    def __init__(self, population=10, sigma=0.1):
        self.population = population
        self.sigma = sigma
        super().__init__()

    def setup_optimization(self):
        x0 = self.tuner.flatten_params()

        options = {"popsize": self.population, "seed": np.random.randint(0, 10000), "tolfun": 1e-5, "tolfunhist": 1e-6}
        self.optim = cma.CMAEvolutionStrategy(x0=x0, sigma0=self.sigma, inopts=options)

    def optimize_step(self):
        params = self.optim.ask()
        # convert params for use

        cost_per_param = []
        all_rollouts = []
        for param in params:
            self.tuner.unflatten_params(param)
            res = self.tuner.evaluate_fn()
            cost_per_param.append(res.costs.mean().cpu().numpy())
            all_rollouts.append(res.rollouts)

        cost_per_param = np.array(cost_per_param)
        self.optim.tell(params, cost_per_param)

        best_param = self.optim.best.x
        self.tuner.unflatten_params(best_param)
        res = self.tuner.evaluate_fn()
        return res


class TunableParameter(abc.ABC):
    """A parameter that can be tuned by the autotuner. Holds references to the object that defines its actual value."""

    @staticmethod
    @abc.abstractmethod
    def name():
        """Get the name of the parameter"""

    @abc.abstractmethod
    def dim(self):
        """Get the dimension of the parameter"""

    @abc.abstractmethod
    def get_current_parameter_value(self):
        """Get the current underlying value of the parameter"""

    @abc.abstractmethod
    def ensure_valid_value(self, value):
        """Return a validated parameter value as close in intent as the input value as possible"""

    @abc.abstractmethod
    def apply_parameter_value(self, value):
        """Apply the parameter value to the underlying object"""

    @abc.abstractmethod
    def attach_to_state(self, state: dict):
        """Reattach/reinitialize the parameter to a new internal state. This should be similar to a call to __init__"""

    def get_parameter_value_from_config(self, config):
        """Get the serialized value of the parameter from a config dictionary, where each name is a scalar"""
        return config[self.name()]

    def get_config_from_parameter_value(self, value):
        """Reverse of the above method, get a config dictionary from a parameter value"""
        return {self.name(): value}


class MPPIParameter(TunableParameter, abc.ABC):
    def __init__(self, mppi: MPPI, dim=None):
        self.mppi = mppi
        self._dim = dim
        if self.mppi is not None:
            self.d = self.mppi.d
            self.dtype = self.mppi.dtype
            if dim is None:
                self._dim = self.mppi.nu

    def attach_to_state(self, state: dict):
        self.mppi = state['mppi']
        self.d = self.mppi.d
        self.dtype = self.mppi.dtype


class SigmaParameter(MPPIParameter):
    eps = 0.0001

    @staticmethod
    def name():
        return 'sigma'

    def dim(self):
        return self._dim

    def get_current_parameter_value(self):
        return torch.cat([self.mppi.noise_sigma[i][i].view(1) for i in range(self.dim())])

    def ensure_valid_value(self, value):
        sigma = ensure_tensor(self.d, self.dtype, value)
        sigma[sigma < self.eps] = self.eps
        return sigma

    def apply_parameter_value(self, value):
        sigma = self.ensure_valid_value(value)
        self.mppi.noise_sigma = torch.diag(sigma)
        self.mppi.noise_dist = MultivariateNormal(self.mppi.noise_mu, covariance_matrix=self.mppi.noise_sigma)
        self.mppi.noise_sigma_inv = torch.inverse(self.mppi.noise_sigma.detach())

    def get_parameter_value_from_config(self, config):
        return torch.tensor([config[f'{self.name()}{i}'] for i in range(self.dim())], dtype=self.dtype, device=self.d)

    def get_config_from_parameter_value(self, value):
        return {f'{self.name()}{i}': value[i].item() for i in range(self.dim())}


class MuParameter(MPPIParameter):
    @staticmethod
    def name():
        return 'mu'

    def dim(self):
        return self._dim

    def get_current_parameter_value(self):
        return self.mppi.noise_mu.clone()

    def ensure_valid_value(self, value):
        mu = ensure_tensor(self.d, self.dtype, value)
        return mu

    def apply_parameter_value(self, value):
        mu = self.ensure_valid_value(value)
        self.mppi.noise_dist = MultivariateNormal(mu, covariance_matrix=self.mppi.noise_sigma)
        self.mppi.noise_sigma_inv = torch.inverse(self.mppi.noise_sigma.detach())

    def get_parameter_value_from_config(self, config):
        return torch.tensor([config[f'{self.name()}{i}'] for i in range(self.dim())], dtype=self.dtype, device=self.d)

    def get_config_from_parameter_value(self, value):
        return {f'{self.name()}{i}': value[i].item() for i in range(self.dim())}


class LambdaParameter(MPPIParameter):
    eps = 0.0001

    @staticmethod
    def name():
        return 'lambda'

    def dim(self):
        return 1

    def get_current_parameter_value(self):
        return self.mppi.lambda_

    def ensure_valid_value(self, value):
        if torch.is_tensor(value) or isinstance(value, np.ndarray):
            value = value[0]
        v = max(value, self.eps)
        return v

    def apply_parameter_value(self, value):
        v = self.ensure_valid_value(value)
        self.mppi.lambda_ = v


class HorizonParameter(MPPIParameter):
    @staticmethod
    def name():
        return 'horizon'

    def dim(self):
        return 1

    def get_current_parameter_value(self):
        return self.mppi.T

    def ensure_valid_value(self, value):
        if torch.is_tensor(value) or isinstance(value, np.ndarray):
            value = value[0]
        v = max(round(value), 1)
        return v

    def apply_parameter_value(self, value):
        v = self.ensure_valid_value(value)
        self.mppi.change_horizon(v)


class Autotune:
    """Tune selected hyperparameters using state-of-the-art optimizers on an evaluation function.
    Subclass to define other parameters to optimize over such as terminal cost scaling. 
    See tests/auto_tune_parameters.py for an example evaluate_fn
    """
    eps = 0.0001

    def __init__(self, params_to_tune: typing.Sequence[TunableParameter],
                 evaluate_fn: typing.Callable[[], EvaluationResult],
                 reload_state_fn: typing.Callable[[], dict] = None,
                 optimizer=CMAESOpt()):
        """

        :param params_to_tune: sequence of tunable parameters
        :param evaluate_fn: function that returns an EvaluationResult that we want to minimize
        :param reload_state_fn: function that returns a dictionary of state to reattach to the parameters
        :param optimizer: optimizer that searches in the parameter space
        """
        self.evaluate_fn = evaluate_fn
        self.reload_state_fn = reload_state_fn

        self.params = params_to_tune
        self.optim = optimizer
        self.optim.tuner = self
        self.results = []

        self.attach_parameters()
        self.optim.setup_optimization()

    def optimize_step(self) -> EvaluationResult:
        res = self.optim.optimize_step()
        res = self.log_current_result(res)
        return res

    def optimize_all(self, iterations) -> EvaluationResult:
        res = self.optim.optimize_all(iterations)
        res = self.log_current_result(res)
        return res

    def get_best_result(self) -> EvaluationResult:
        return min(self.results, key=lambda res: res.costs.mean().item())

    def log_current_result(self, res: EvaluationResult):
        with torch.no_grad():
            iteration = len(self.results)
            kv = self.get_parameter_values(self.params)
            res = res._replace(iteration=iteration,
                               params={k: v.detach().clone() if torch.is_tensor(v) else v for k, v in
                                       kv.items()})
            logger.info(f"i:{iteration} cost: {res.costs.mean().item()} params:{kv}")
            self.results.append(res)
        return res

    def get_parameter_values(self, params_to_tune: typing.Sequence[TunableParameter]):
        # take on the assigned values to the MPPI
        return {p.name(): p.get_current_parameter_value() for p in params_to_tune}

    def flatten_params(self):
        x = []
        kv = self.get_parameter_values(self.params)
        # TODO ensure this is the same order as define and unflatten
        for k, v in kv.items():
            if torch.is_tensor(v):
                x.append(v.detach().cpu().numpy())
            else:
                x.append([v])
        x = np.concatenate(x)
        return x

    def unflatten_params(self, x, apply=True):
        # have to be in the same order as the flattening
        param_values = {}
        i = 0
        for p in self.params:
            raw_value = x[i:i + p.dim()]
            param_values[p.name()] = p.ensure_valid_value(raw_value)
            i += p.dim()
        if apply:
            self.apply_parameters(param_values)
        return param_values

    def apply_parameters(self, param_values):
        for p in self.params:
            p.apply_parameter_value(param_values[p.name()])

    def attach_parameters(self):
        """Attach parameters to any underlying state they require In most cases the parameters are defined already
        attached to whatever state it needs, e.g. the MPPI controller object for changing the parameter values.
        However, there are cases where the full state is not serializable, e.g. when using a multiprocessing pool
        and so we pass only the information required to load the state. We then must load the state and reattach
        the parameters to the state each training iteration."""
        if self.reload_state_fn is not None:
            state = self.reload_state_fn()
            for p in self.params:
                p.attach_to_state(state)

    def config_to_params(self, config):
        """Configs are param dictionaries where each must be a scalar"""
        return {p.name(): p.get_parameter_value_from_config(config) for p in self.params}



import logging
import time
import typing

import torch
from torch.distributions.multivariate_normal import MultivariateNormal
from arm_pytorch_utilities import handle_batch_input
from functorch import vmap

logger = logging.getLogger(__name__)


def _ensure_non_zero(cost, beta, factor):
    return torch.exp(-factor * (cost - beta))


class SpecificActionSampler:
    def __init__(self):
        self.start_idx = 0
        self.end_idx = 0
        self.slice = slice(0, 0)

    def sample_trajectories(self, state, info):
        raise NotImplementedError

    def specific_dynamics(self, next_state, state, action, t):
        """Handle dynamics in a specific way for the specific action sampler; defaults to using default dynamics"""
        return next_state

    def register_sample_start_end(self, start_idx, end_idx):
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.slice = slice(start_idx, end_idx)


class MPPI():
    """
    Model Predictive Path Integral control
    This implementation batch samples the trajectories and so scales well with the number of samples K.

    Implemented according to algorithm 2 in Williams et al., 2017
    'Information Theoretic MPC for Model-Based Reinforcement Learning',
    based off of https://github.com/ferreirafabio/mppi_pendulum
    """

    def __init__(self, dynamics, running_cost, nx, noise_sigma, num_samples=100, horizon=15, device="cpu",
                 terminal_state_cost=None,
                 lambda_=1.,
                 noise_mu=None,
                 u_min=None,
                 u_max=None,
                 u_init=None,
                 U_init=None,
                 u_scale=1,
                 u_per_command=1,
                 step_dependent_dynamics=False,
                 rollout_samples=1,
                 rollout_var_cost=0,
                 rollout_var_discount=0.95,
                 sample_null_action=False,
                 specific_action_sampler: typing.Optional[SpecificActionSampler] = None,
                 noise_abs_cost=False):
        """
        :param dynamics: function(state, action) -> next_state (K x nx) taking in batch state (K x nx) and action (K x nu)
        :param running_cost: function(state, action) -> cost (K) taking in batch state and action (same as dynamics)
        :param nx: state dimension
        :param noise_sigma: (nu x nu) control noise covariance (assume v_t ~ N(u_t, noise_sigma))
        :param num_samples: K, number of trajectories to sample
        :param horizon: T, length of each trajectory
        :param device: pytorch device
        :param terminal_state_cost: function(state) -> cost (K x 1) taking in batch state
        :param lambda_: temperature, positive scalar where larger values will allow more exploration
        :param noise_mu: (nu) control noise mean (used to bias control samples); defaults to zero mean
        :param u_min: (nu) minimum values for each dimension of control to pass into dynamics
        :param u_max: (nu) maximum values for each dimension of control to pass into dynamics
        :param u_init: (nu) what to initialize new end of trajectory control to be; defeaults to zero
        :param U_init: (T x nu) initial control sequence; defaults to noise
        :param step_dependent_dynamics: whether the passed in dynamics needs horizon step passed in (as 3rd arg)
        :param rollout_samples: M, number of state trajectories to rollout for each control trajectory
            (should be 1 for deterministic dynamics and more for models that output a distribution)
        :param rollout_var_cost: Cost attached to the variance of costs across trajectory rollouts
        :param rollout_var_discount: Discount of variance cost over control horizon
        :param sample_null_action: Whether to explicitly sample a null action (bad for starting in a local minima)
        :param specific_action_sampler: Function to explicitly sample actions to use instead of sampling from noise from
            nominal trajectory, may output a number of action trajectories fewer than horizon
        :param noise_abs_cost: Whether to use the absolute value of the action noise to avoid bias when all states have the same cost
        """
        self.d = device
        self.dtype = noise_sigma.dtype
        self.K = num_samples  # N_SAMPLES
        self.T = horizon  # TIMESTEPS

        # dimensions of state and control
        self.nx = nx
        self.nu = 1 if len(noise_sigma.shape) == 0 else noise_sigma.shape[0]
        self.lambda_ = lambda_

        if noise_mu is None:
            noise_mu = torch.zeros(self.nu, dtype=self.dtype)

        if u_init is None:
            u_init = torch.zeros_like(noise_mu)

        # handle 1D edge case
        if self.nu == 1:
            noise_mu = noise_mu.view(-1)
            noise_sigma = noise_sigma.view(-1, 1)

        # bounds
        self.u_min = u_min
        self.u_max = u_max
        self.u_scale = u_scale
        self.u_per_command = u_per_command
        # make sure if any of them is specified, both are specified
        if self.u_max is not None and self.u_min is None:
            if not torch.is_tensor(self.u_max):
                self.u_max = torch.tensor(self.u_max)
            self.u_min = -self.u_max
        if self.u_min is not None and self.u_max is None:
            if not torch.is_tensor(self.u_min):
                self.u_min = torch.tensor(self.u_min)
            self.u_max = -self.u_min
        if self.u_min is not None:
            self.u_min = self.u_min.to(device=self.d)
            self.u_max = self.u_max.to(device=self.d)

        self.noise_mu = noise_mu.to(self.d)
        self.noise_sigma = noise_sigma.to(self.d)
        self.noise_sigma_inv = torch.inverse(self.noise_sigma)
        self.noise_dist = MultivariateNormal(self.noise_mu, covariance_matrix=self.noise_sigma)
        # T x nu control sequence
        self.U = U_init
        self.u_init = u_init.to(self.d)

        if self.U is None:
            self.U = self.noise_dist.sample((self.T,))

        self.step_dependency = step_dependent_dynamics
        self.F = dynamics
        self.running_cost = running_cost
        self.terminal_state_cost = terminal_state_cost
        self.sample_null_action = sample_null_action
        self.specific_action_sampler = specific_action_sampler
        self.noise_abs_cost = noise_abs_cost
        self.state = None
        self.info = None

        # handling dynamics models that output a distribution (take multiple trajectory samples)
        self.M = rollout_samples
        self.rollout_var_cost = rollout_var_cost
        self.rollout_var_discount = rollout_var_discount

        # sampled results from last command
        self.cost_total = None
        self.cost_total_non_zero = None
        self.omega = None
        self.states = None
        self.actions = None

    def get_params(self):
        return f"K={self.K} T={self.T} M={self.M} lambda={self.lambda_} noise_mu={self.noise_mu.cpu().numpy()} noise_sigma={self.noise_sigma.cpu().numpy()}".replace(
            "\n", ",")

    @handle_batch_input(n=2)
    def _dynamics(self, state, u, t):
        return self.F(state, u, t) if self.step_dependency else self.F(state, u)

    @handle_batch_input(n=2)
    def _running_cost(self, state, u, t):
        return self.running_cost(state, u, t) if self.step_dependency else self.running_cost(state, u)

    def get_action_sequence(self):
        return self.U

    def shift_nominal_trajectory(self):
        """
        Shift the nominal trajectory forward one step
        """
        # shift command 1 time step
        self.U = torch.roll(self.U, -1, dims=0)
        self.U[-1] = self.u_init

    def command(self, state, shift_nominal_trajectory=True, info=None):
        """
        :param state: (nx) or (K x nx) current state, or samples of states (for propagating a distribution of states)
        :param shift_nominal_trajectory: Whether to roll the nominal trajectory forward one step. This should be True
            if the command is to be executed. If the nominal trajectory is to be refined then it should be False.
        :param info: Optional dictionary to store context information
        :returns action: (nu) best action
        """
        self.info = info
        if shift_nominal_trajectory:
            self.shift_nominal_trajectory()

        return self._command(state)

    def _compute_weighting(self, cost_total):
        beta = torch.min(cost_total)
        self.cost_total_non_zero = _ensure_non_zero(cost_total, beta, 1 / self.lambda_)
        eta = torch.sum(self.cost_total_non_zero)
        self.omega = (1. / eta) * self.cost_total_non_zero
        return self.omega

    def _command(self, state):
        if not torch.is_tensor(state):
            state = torch.tensor(state)
        self.state = state.to(dtype=self.dtype, device=self.d)
        cost_total = self._compute_total_cost_batch()

        self._compute_weighting(cost_total)
        perturbations = torch.sum(self.omega.view(-1, 1, 1) * self.noise, dim=0)

        self.U = self.U + perturbations
        action = self.get_action_sequence()[:self.u_per_command]
        # reduce dimensionality if we only need the first command
        if self.u_per_command == 1:
            action = action[0]
        return action

    def change_horizon(self, horizon):
        if horizon < self.U.shape[0]:
            # truncate trajectory
            self.U = self.U[:horizon]
        elif horizon > self.U.shape[0]:
            # extend with u_init
            self.U = torch.cat((self.U, self.u_init.repeat(horizon - self.U.shape[0], 1)))
        self.T = horizon

    def reset(self):
        """
        Clear controller state after finishing a trial
        """
        self.U = self.noise_dist.sample((self.T,))

    def _compute_rollout_costs(self, perturbed_actions):
        K, T, nu = perturbed_actions.shape
        assert nu == self.nu

        cost_total = torch.zeros(K, device=self.d, dtype=self.dtype)
        cost_samples = cost_total.repeat(self.M, 1)
        cost_var = torch.zeros_like(cost_total)

        # allow propagation of a sample of states (ex. to carry a distribution), or to start with a single state
        if self.state.shape == (K, self.nx):
            state = self.state
        else:
            state = self.state.view(1, -1).repeat(K, 1)

        # rollout action trajectory M times to estimate expected cost
        state = state.repeat(self.M, 1, 1)

        states = []
        actions = []
        for t in range(T):
            u = self.u_scale * perturbed_actions[:, t].repeat(self.M, 1, 1)
            next_state = self._dynamics(state, u, t)
            # potentially handle dynamics in a specific way for the specific action sampler
            next_state = self._sample_specific_dynamics(next_state, state, u, t)
            state = next_state
            c = self._running_cost(state, u, t)
            cost_samples = cost_samples + c
            if self.M > 1:
                cost_var += c.var(dim=0) * (self.rollout_var_discount ** t)

            # Save total states/actions
            states.append(state)
            actions.append(u)

        # Actions is K x T x nu
        # States is K x T x nx
        actions = torch.stack(actions, dim=-2)
        states = torch.stack(states, dim=-2)

        # action perturbation cost
        if self.terminal_state_cost:
            c = self.terminal_state_cost(states, actions)
            cost_samples = cost_samples + c
        cost_total = cost_total + cost_samples.mean(dim=0)
        cost_total = cost_total + cost_var * self.rollout_var_cost
        return cost_total, states, actions

    def _compute_perturbed_action_and_noise(self):
        # parallelize sampling across trajectories
        # resample noise each time we take an action
        noise = self.noise_dist.rsample((self.K, self.T))
        # broadcast own control to noise over samples; now it's K x T x nu
        perturbed_action = self.U + noise
        perturbed_action = self._sample_specific_actions(perturbed_action)
        # naively bound control
        self.perturbed_action = self._bound_action(perturbed_action)
        # bounded noise after bounding (some got cut off, so we don't penalize that in action cost)
        self.noise = self.perturbed_action - self.U

    def _sample_specific_actions(self, perturbed_action):
        # specific sampling of actions (encoding trajectory prior and domain knowledge to create biases)
        i = 0
        if self.sample_null_action:
            perturbed_action[i] = 0
            i += 1
        if self.specific_action_sampler is not None:
            actions = self.specific_action_sampler.sample_trajectories(self.state, self.info)
            # check how long it is
            actions = actions.reshape(-1, self.T, self.nu)
            perturbed_action[i:i + actions.shape[0]] = actions
            self.specific_action_sampler.register_sample_start_end(i, i + actions.shape[0])
            i += actions.shape[0]
        return perturbed_action

    def _sample_specific_dynamics(self, next_state, state, u, t):
        if self.specific_action_sampler is not None:
            next_state = self.specific_action_sampler.specific_dynamics(next_state, state, u, t)
        return next_state

    def _compute_total_cost_batch(self):
        self._compute_perturbed_action_and_noise()
        if self.noise_abs_cost:
            action_cost = self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv
            # NOTE: The original paper does self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv, but this biases
            # the actions with low noise if all states have the same cost. With abs(noise) we prefer actions close to the
            # nomial trajectory.
        else:
            action_cost = self.lambda_ * self.noise @ self.noise_sigma_inv  # Like original paper

        rollout_cost, self.states, actions = self._compute_rollout_costs(self.perturbed_action)
        self.actions = actions / self.u_scale

        # action perturbation cost
        perturbation_cost = torch.sum(self.U * action_cost, dim=(1, 2))
        self.cost_total = rollout_cost + perturbation_cost
        return self.cost_total

    def _bound_action(self, action):
        if self.u_max is not None:
            return torch.max(torch.min(action, self.u_max), self.u_min)
        return action

    def _slice_control(self, t):
        return slice(t * self.nu, (t + 1) * self.nu)

    def get_rollouts(self, state, num_rollouts=1, U=None):
        """
            :param state: either (nx) vector or (num_rollouts x nx) for sampled initial states
            :param num_rollouts: Number of rollouts with same action sequence - for generating samples with stochastic
                                 dynamics
            :returns states: num_rollouts x T x nx vector of trajectories

        """
        state = state.view(-1, self.nx)
        if state.size(0) == 1:
            state = state.repeat(num_rollouts, 1)

        if U is None:
            U = self.get_action_sequence()
        T = U.shape[0]
        states = torch.zeros((num_rollouts, T + 1, self.nx), dtype=U.dtype, device=U.device)
        states[:, 0] = state
        for t in range(T):
            next_state = self._dynamics(states[:, t].view(num_rollouts, -1),
                                        self.u_scale * U[t].tile(num_rollouts, 1), t)
            # dynamics may augment state; here we just take the first nx dimensions
            states[:, t + 1] = next_state[:, :self.nx]

        return states[:, 1:]


class SMPPI(MPPI):
    """Smooth MPPI by lifting the control space and penalizing the change in action from
    https://arxiv.org/pdf/2112.09988
    """

    def __init__(self, *args, w_action_seq_cost=1., delta_t=1., U_init=None, action_min=None, action_max=None,
                 **kwargs):
        self.w_action_seq_cost = w_action_seq_cost
        self.delta_t = delta_t

        super().__init__(*args, U_init=U_init, **kwargs)

        # these are the actual commanded actions, which is now no longer directly sampled
        self.action_min = action_min
        self.action_max = action_max
        if self.action_min is not None and self.action_max is None:
            if not torch.is_tensor(self.action_min):
                self.action_min = torch.tensor(self.action_min)
            self.action_max = -self.action_min
        if self.action_max is not None and self.action_min is None:
            if not torch.is_tensor(self.action_max):
                self.action_max = torch.tensor(self.action_max)
            self.action_min = -self.action_max
        if self.action_min is not None:
            self.action_min = self.action_min.to(device=self.d)
            self.action_max = self.action_max.to(device=self.d)

        # this smooth formulation works better if control starts from 0
        if U_init is None:
            self.action_sequence = torch.zeros_like(self.U)
        else:
            self.action_sequence = U_init
        self.U = torch.zeros_like(self.U)

    def get_params(self):
        return f"{super().get_params()} w={self.w_action_seq_cost} t={self.delta_t}"

    def shift_nominal_trajectory(self):
        self.U = torch.roll(self.U, -1, dims=0)
        self.U[-1] = self.u_init
        self.action_sequence = torch.roll(self.action_sequence, -1, dims=0)
        self.action_sequence[-1] = self.action_sequence[-2]  # add T-1 action to T

    def get_action_sequence(self):
        return self.action_sequence

    def reset(self):
        self.U = torch.zeros_like(self.U)
        self.action_sequence = torch.zeros_like(self.U)

    def change_horizon(self, horizon):
        if horizon < self.U.shape[0]:
            # truncate trajectory
            self.U = self.U[:horizon]
            self.action_sequence = self.action_sequence[:horizon]
        elif horizon > self.U.shape[0]:
            # extend with u_init
            extend_for = horizon - self.U.shape[0]
            self.U = torch.cat((self.U, self.u_init.repeat(extend_for, 1)))
            self.action_sequence = torch.cat((self.action_sequence, self.action_sequence[-1].repeat(extend_for, 1)))
        self.T = horizon

    def _bound_d_action(self, control):
        if self.u_max is not None:
            return torch.max(torch.min(control, self.u_max), self.u_min)  # action
        return control

    def _bound_action(self, action):
        if self.action_max is not None:
            return torch.max(torch.min(action, self.action_max), self.action_min)
        return action

    def _command(self, state):
        if not torch.is_tensor(state):
            state = torch.tensor(state)
        self.state = state.to(dtype=self.dtype, device=self.d)
        cost_total = self._compute_total_cost_batch()

        self._compute_weighting(cost_total)
        perturbations = torch.sum(self.omega.view(-1, 1, 1) * self.noise, dim=0)

        self.U = self.U + perturbations
        # U is now the lifted control space, so we integrate it
        self.action_sequence += self.U * self.delta_t

        action = self.get_action_sequence()[:self.u_per_command]
        # reduce dimensionality if we only need the first command
        if self.u_per_command == 1:
            action = action[0]
        return action

    def _compute_perturbed_action_and_noise(self):
        # parallelize sampling across trajectories
        # resample noise each time we take an action
        noise = self.noise_dist.rsample((self.K, self.T))
        # broadcast own control to noise over samples; now it's K x T x nu
        perturbed_control = self.U + noise
        # naively bound control
        self.perturbed_control = self._bound_d_action(perturbed_control)
        # bounded noise after bounding (some got cut off, so we don't penalize that in action cost)
        self.perturbed_action = self.action_sequence + perturbed_control * self.delta_t
        self.perturbed_action = self._sample_specific_actions(self.perturbed_action)
        self.perturbed_action = self._bound_action(self.perturbed_action)

        self.noise = (self.perturbed_action - self.action_sequence) / self.delta_t - self.U

    def _compute_total_cost_batch(self):
        self._compute_perturbed_action_and_noise()
        if self.noise_abs_cost:
            action_cost = self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv
            # NOTE: The original paper does self.lambda_ * torch.abs(self.noise) @ self.noise_sigma_inv, but this biases
            # the actions with low noise if all states have the same cost. With abs(noise) we prefer actions close to the
            # nomial trajectory.
        else:
            action_cost = self.lambda_ * self.noise @ self.noise_sigma_inv  # Like original paper

        # action difference as cost
        action_diff = self.u_scale * torch.diff(self.perturbed_action, dim=-2)
        action_smoothness_cost = torch.sum(torch.square(action_diff), dim=(1, 2))
        # handle non-homogeneous action sequence cost
        action_smoothness_cost *= self.w_action_seq_cost

        rollout_cost, self.states, actions = self._compute_rollout_costs(self.perturbed_action)
        self.actions = actions / self.u_scale

        # action perturbation cost
        perturbation_cost = torch.sum(self.U * action_cost, dim=(1, 2))
        self.cost_total = rollout_cost + perturbation_cost + action_smoothness_cost
        return self.cost_total


class TimeKernel:
    """Kernel acting on the time dimension of trajectories for use in interpolation and smoothing"""

    def __call__(self, t, tk):
        raise NotImplementedError


class RBFKernel(TimeKernel):
    def __init__(self, sigma=1):
        self.sigma = sigma

    def __repr__(self):
        return f"RBFKernel(sigma={self.sigma})"

    def __call__(self, t, tk):
        d = torch.sum((t[:, None] - tk) ** 2, dim=-1)
        k = torch.exp(-d / (1e-8 + 2 * self.sigma ** 2))
        return k


class KMPPI(MPPI):
    """MPPI with kernel interpolation of control points for smoothing"""

    def __init__(self, *args, num_support_pts=None, kernel: TimeKernel = RBFKernel(), **kwargs):
        super().__init__(*args, **kwargs)
        self.num_support_pts = num_support_pts or self.T // 2
        # control points to be sampled
        self.theta = torch.zeros((self.num_support_pts, self.nu), dtype=self.dtype, device=self.d)
        self.Tk = None
        self.Hs = None
        # interpolation kernel
        self.interpolation_kernel = kernel
        self.intp_krnl = None
        self.prepare_vmap_interpolation()

    def get_params(self):
        return f"{super().get_params()} num_support_pts={self.num_support_pts} kernel={self.interpolation_kernel}"

    def reset(self):
        super().reset()
        self.theta.zero_()

    def shift_nominal_trajectory(self):
        super().shift_nominal_trajectory()
        self.theta, _ = self.do_kernel_interpolation(self.Tk[0] + 1, self.Tk[0], self.theta)

    def do_kernel_interpolation(self, t, tk, c):
        K = self.interpolation_kernel(t.unsqueeze(-1), tk.unsqueeze(-1))
        Ktktk = self.interpolation_kernel(tk.unsqueeze(-1), tk.unsqueeze(-1))
        # print(K.shape, Ktktk.shape)
        # row normalize K
        # K = K / K.sum(dim=1).unsqueeze(1)

        # KK = K @ torch.inverse(Ktktk)
        KK = torch.linalg.solve(Ktktk, K, left=False)

        return torch.matmul(KK, c), K

    def prepare_vmap_interpolation(self):
        self.Tk = torch.linspace(0, self.T - 1, int(self.num_support_pts), device=self.d, dtype=self.dtype).unsqueeze(
            0).repeat(self.K, 1)
        self.Hs = torch.linspace(0, self.T - 1, int(self.T), device=self.d, dtype=self.dtype).unsqueeze(0).repeat(
            self.K, 1)
        self.intp_krnl = vmap(self.do_kernel_interpolation)

    def deparameterize_to_trajectory_single(self, theta):
        return self.do_kernel_interpolation(self.Hs[0], self.Tk[0], theta)

    def deparameterize_to_trajectory_batch(self, theta):
        assert theta.shape == (self.K, self.num_support_pts, self.nu)
        return self.intp_krnl(self.Hs, self.Tk, theta)

    def _compute_perturbed_action_and_noise(self):
        # parallelize sampling across trajectories
        # resample noise each time we take an action
        noise = self.noise_dist.rsample((self.K, self.num_support_pts))
        perturbed_control_pts = self.theta + noise
        # control points in the same space as control and should be bounded
        perturbed_control_pts = self._bound_action(perturbed_control_pts)
        self.noise_theta = perturbed_control_pts - self.theta
        perturbed_action, _ = self.deparameterize_to_trajectory_batch(perturbed_control_pts)
        perturbed_action = self._sample_specific_actions(perturbed_action)
        # naively bound control
        self.perturbed_action = self._bound_action(perturbed_action)
        # bounded noise after bounding (some got cut off, so we don't penalize that in action cost)
        self.noise = self.perturbed_action - self.U

    def _command(self, state):
        if not torch.is_tensor(state):
            state = torch.tensor(state)
        self.state = state.to(dtype=self.dtype, device=self.d)
        cost_total = self._compute_total_cost_batch()

        self._compute_weighting(cost_total)
        perturbations = torch.sum(self.omega.view(-1, 1, 1) * self.noise_theta, dim=0)

        self.theta = self.theta + perturbations
        self.U, _ = self.deparameterize_to_trajectory_single(self.theta)

        action = self.get_action_sequence()[:self.u_per_command]
        # reduce dimensionality if we only need the first command
        if self.u_per_command == 1:
            action = action[0]
        return action


def run_mppi(mppi, env, retrain_dynamics, retrain_after_iter=50, iter=1000, render=True):
    dataset = torch.zeros((retrain_after_iter, mppi.nx + mppi.nu), dtype=mppi.U.dtype, device=mppi.d)
    total_reward = 0
    for i in range(iter):
        state = env.unwrapped.state.copy()
        command_start = time.perf_counter()
        action = mppi.command(state)
        elapsed = time.perf_counter() - command_start
        res = env.step(action.cpu().numpy())
        s, r = res[0], res[1]
        total_reward += r
        logger.debug("action taken: %.4f cost received: %.4f time taken: %.5fs", action, -r, elapsed)
        if render:
            env.render()

        di = i % retrain_after_iter
        if di == 0 and i > 0:
            retrain_dynamics(dataset)
            # don't have to clear dataset since it'll be overridden, but useful for debugging
            dataset.zero_()
        dataset[di, :mppi.nx] = torch.tensor(state, dtype=mppi.U.dtype)
        dataset[di, mppi.nx:] = action
    return total_reward, dataset



import abc
import numpy as np
import torch.cuda

# pip install "ray[tune]" bayesian-optimization hyperopt
from ray import tune
from ray import train

from pytorch_mppi import autotune
from ray.tune.search.hyperopt import HyperOptSearch


class GlobalTunableParameter(autotune.TunableParameter, abc.ABC):
    def __init__(self, search_space):
        self.search_space = search_space

    @abc.abstractmethod
    def total_search_space(self) -> dict:
        """Return the potentially multidimensional search space for this parameter, which is a dictionary mapping
        each of the parameter's corresponding config names to a search space."""

    def get_linearized_search_space_value(self, param_values):
        if self.dim() == 1:
            return [self._linearize_space_value(self.search_space, param_values[self.name()])]
        return [self._linearize_space_value(self.search_space, param_values[f"{self.name()}"][i].item()) for i in
                range(self.dim())]

    @staticmethod
    def linearize_search_space(space):
        # tune doesn't have public API for type checking samplers
        sampler = space.get_sampler()
        if hasattr(sampler, 'base'):
            b = np.log(sampler.base)
            return np.log(space.lower) / b, np.log(space.upper) / b
        return space.lower, space.upper

    @staticmethod
    def _linearize_space_value(space, v):
        # tune doesn't have public API for type checking samplers
        sampler = space.get_sampler()
        # log
        if hasattr(sampler, 'base'):
            b = np.log(sampler.base)
            return np.log(v) / b
        # quantized
        if hasattr(sampler, 'q'):
            return np.round(np.divide(v, sampler.q)) * sampler.q
        return v


class SigmaGlobalParameter(autotune.SigmaParameter, GlobalTunableParameter):
    def __init__(self, *args, search_space=tune.loguniform(1e-4, 1e2), **kwargs):
        super().__init__(*args, **kwargs)
        GlobalTunableParameter.__init__(self, search_space)

    def total_search_space(self) -> dict:
        return {f"{self.name()}{i}": self.search_space for i in range(self.dim())}


class MuGlobalParameter(autotune.MuParameter, GlobalTunableParameter):
    def __init__(self, *args, search_space=tune.uniform(-1, 1), **kwargs):
        super().__init__(*args, **kwargs)
        GlobalTunableParameter.__init__(self, search_space)

    def total_search_space(self) -> dict:
        return {f"{self.name()}{i}": self.search_space for i in range(self.dim())}


class LambdaGlobalParameter(autotune.LambdaParameter, GlobalTunableParameter):
    def __init__(self, *args, search_space=tune.loguniform(1e-5, 1e3), **kwargs):
        super().__init__(*args, **kwargs)
        GlobalTunableParameter.__init__(self, search_space)

    def total_search_space(self) -> dict:
        return {self.name(): self.search_space}


class HorizonGlobalParameter(autotune.HorizonParameter, GlobalTunableParameter):
    def __init__(self, *args, search_space=tune.randint(1, 50), **kwargs):
        super().__init__(*args, **kwargs)
        GlobalTunableParameter.__init__(self, search_space)

    def total_search_space(self) -> dict:
        return {self.name(): self.search_space}


class AutotuneGlobal(autotune.Autotune):
    def search_space(self):
        space = {}
        for p in self.params:
            assert isinstance(p, GlobalTunableParameter)
            space.update(p.total_search_space())
        return space

    def linearized_search_space(self):
        return {k: GlobalTunableParameter.linearize_search_space(space) for k, space in self.search_space().items()}

    def linearize_params(self, param_values):
        v = []
        for p in self.params:
            assert isinstance(p, GlobalTunableParameter)
            v.extend(p.get_linearized_search_space_value(param_values))
        return np.array(v)

    def initial_value(self):
        init = {}
        param_values = self.get_parameter_values(self.params)
        for p in self.params:
            assert isinstance(p, GlobalTunableParameter)
            init.update(p.get_config_from_parameter_value(param_values[p.name()]))
        return init


class RayOptimizer(autotune.Optimizer):
    def __init__(self, search_alg=HyperOptSearch,
                 default_iterations=100):
        self.iterations = default_iterations
        self.search_alg = search_alg
        self.all_res = None
        super().__init__()

    def setup_optimization(self):
        if not isinstance(self.tuner, AutotuneGlobal):
            raise RuntimeError(f"Ray optimizers require global search space information provided by AutotuneMPPIGlobal")
        space = self.tuner.search_space()
        init = self.tuner.initial_value()

        hyperopt_search = self.search_alg(points_to_evaluate=[init], metric="cost", mode="min")

        trainable_with_resources = tune.with_resources(self.trainable, {"gpu": 1 if torch.cuda.is_available() else 0})
        self.optim = tune.Tuner(
            trainable_with_resources,
            tune_config=tune.TuneConfig(
                num_samples=self.iterations,
                search_alg=hyperopt_search,
                metric="cost",
                mode="min",
            ),
            param_space=space,
        )

    def trainable(self, config):
        self.tuner.attach_parameters()
        self.tuner.apply_parameters(self.tuner.config_to_params(config))
        res = self.tuner.evaluate_fn()
        train.report({'cost': res.costs.mean().item()})

    def optimize_step(self):
        raise RuntimeError("Ray optimizers only allow tuning of all iterations at once")

    def optimize_all(self, iterations):
        self.iterations = iterations
        self.setup_optimization()
        self.all_res = self.optim.fit()
        self.tuner.apply_parameters(self.tuner.config_to_params(self.all_res.get_best_result().config))
        res = self.tuner.evaluate_fn()
        return res
